// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.features;

import imagescience.images.Axes;
import imagescience.images.ByteImage;
import imagescience.images.Coordinates;
import imagescience.images.Dimensions;
import imagescience.images.FloatImage;
import imagescience.images.PixelSizes;
import imagescience.images.ShortImage;
import imagescience.utilities.Messenger;
import imagescience.utilities.Progressor;
import imagescience.utilities.Timer;

/** Provides methods for computing Hessian eigenimages. */
public final class Hessian {

    /** Method for computing Hessian eigenimages of images of type
	<code>ByteImage</code>. See the <code>run</code> method
	for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage[] run(final ByteImage image,
			    final float scale,
			    final boolean absolute) {

	return run(new FloatImage(image),scale,absolute,true);
    }

    /** Method for computing Hessian eigenimages of images of type
	<code>ShortImage</code>. See the <code>run</code> method
	for <code>FloatImage</code> objects for details on the
	parameters and exceptions.
    */
    public FloatImage[] run(final ShortImage image,
			    final float scale,
			    final boolean absolute) {

	return run(new FloatImage(image),scale,absolute,true);
    }

    /** Method for computing Hessian eigenimages of images of type
	<code>FloatImage</code>.

	@param image the input image for which Hessian eigenimages need
	to be computed.<br><ul>If the size of the image in the
	z-dimension equals one, this method will compute, for every
	pixel, the two-dimensional (2D) Hessian and its two
	eigenvalues. Otherwise it will compute for every pixel the full
	three-dimensional (3D) Hessian and its three eigenvalues. These
	computations are applied to every volume in the fourth
	dimension.</ul>

	@param scale the smoothing scale at which the required image
	derivatives are computed.<br><ul> The scale is equal to the
	standard deviation of the Gaussian kernel used for
	differentiation and must be larger than zero. In order to
	enforce physical isotropy, for each dimension, the scale is
	divided by the pixel size of the image in that dimension.</ul>

	@param absolute determines whether or not eigenvalues are
	compared in absolute sense.

	@param useimage indicates whether or not the contents of the
	input image may be altered by the method.<br><ul>If this
	parameter is <code>true</code>, the method may use the input
	image rather than a newly created image to store intermediate
	results, thereby saving memory. If it is <code>false</code>, the
	input image is left unaltered.</ul>
	
	@return an array containing the eigenimages.<br><ul>If only the
	two-dimensional (2D) Hessian and its two eigenvalues were
	computed for every pixel, the returned array contains two
	eigenimages: at index 0 the image with, for every pixel, the
	largest (absolute) eigenvalue; at index 1 the image with, for
	every pixel, the smallest (absolute) eigenvalue. If the full
	three-dimensional (3D) Hessian and its three eigenvalues were
	computed for every pixel, the returned array contains three
	eigenimages: at index 0 the image with, for every pixel, the
	largest (absolute) eigenvalue; at index 1 the image with, for
	every pixel, the middle (absolute) eigenvalue; at index 2 the
	image with, for every pixel, the smallest (absolute)
	eigenvalue.</ul>

	@exception IllegalArgumentException if the scale is negative.

	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public FloatImage[] run(final FloatImage image,
			    final float scale,
			    final boolean absolute,
			    final boolean useimage) {

	Features.printVersion("Hessian");

	final Timer tmr = new Timer(); tmr.start();

	// Initialize:
	msg.debug("Checking arguments for Hessian computation");
	if (scale < 0.0f) throw new IllegalArgumentException("Negative smoothing scale");

	final Dimensions dims = image.dimensions();
	msg.debug("X-size of input image: " + dims.x);
	msg.debug("Y-size of input image: " + dims.y);
	msg.debug("Z-size of input image: " + dims.z);
	msg.debug("T-size of input image: " + dims.t);

	final PixelSizes ps = image.pixelsizes();
	msg.debug("Pixel size in x-dimension: " + ps.x);
	msg.debug("Pixel size in y-dimension: " + ps.y);
	msg.debug("Pixel size in z-dimension: " + ps.z);

	FloatImage[] eigenimages = null;

	// Compute Hessian matrix (2D or 3D) and eigenimages:
	if (dims.z == 1) { // 2D case

	    // Compute the components of the Hessian:
	    final boolean bDebug = msg.debug();
	    FloatImage Hxx, Hxy, Hyy;
	    if (Features.sepSmooth()) {
		msg.debug("Seperating smoothing and differentiation");
		msg.debug("Gaussian smoothing at scale "+scale);
		msg.debug(false);
		final FloatImage smoothImage = useimage ? image : image.duplicate();
		differ.smooth(smoothImage,scale,false);
		msg.debug(bDebug);
		msg.debug("Computing Hxx");
		msg.debug(false);
		Hxx = differ.fdiff(smoothImage,0.0f,2,0,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Hxy");
		msg.debug(false);
		Hxy = differ.fdiff(smoothImage,0.0f,1,1,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Hyy");
		msg.debug(false);
		Hyy = differ.fdiff(smoothImage,0.0f,0,2,0,false);
		msg.debug(bDebug);
	    } else {
		msg.debug("Combining smoothing and differentiation");
		msg.debug("Gaussian differentiation at scale "+scale);
		msg.debug("Computing Hxx");
		msg.debug(false);
		Hxx = differ.gauss(image,scale,2,0,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Hxy");
		msg.debug(false);
		Hxy = differ.gauss(image,scale,1,1,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Hyy");
		msg.debug(false);
		Hyy = differ.gauss(image,scale,0,2,0,useimage?false:true);
		msg.debug(bDebug);
	    }

	    // Compute the eigenimages: (note that in order to minimize
	    // memory usage, Hxx and Hyy are re-used)
	    msg.debug("Computing 2D eigenimages");
	    final Progressor pgs = new Progressor("Computing eigenimages");
	    pgs.steps(dims.t*dims.z*dims.y);
	    Hxx.axes(Axes.X); Hxy.axes(Axes.X); Hyy.axes(Axes.X);
	    final float[] ahxx = new float[dims.x];
	    final float[] ahxy = new float[dims.x];
	    final float[] ahyy = new float[dims.x];
	    final Coordinates coords = new Coordinates();

	    if (absolute) {
		msg.debug("Comparing and storing absolute eigenvalues");
		for (coords.t=0; coords.t<dims.t; ++coords.t)
		    for (coords.y=0; coords.y<dims.y; ++coords.y) {
			Hxx.get(coords,ahxx);
			Hxy.get(coords,ahxy);
			Hyy.get(coords,ahyy);
			for (int x=0; x<dims.x; ++x) {
			    final float b = -(ahxx[x] + ahyy[x]);
			    final float c = ahxx[x]*ahyy[x] - ahxy[x]*ahxy[x];
			    final float q = (float)(-0.5*(b + (b < 0.0f ? -1.0f : 1.0f)*Math.sqrt(b*b - 4*c)));
			    final float absh1 = Math.abs(q);
			    final float absh2 = Math.abs(c/q);
			    if (absh1 > absh2) { ahxx[x] = absh1; ahyy[x] = absh2; }
			    else { ahxx[x] = absh2; ahyy[x] = absh1; }
			}
			Hxx.set(coords,ahxx);
			Hyy.set(coords,ahyy);
			pgs.step();
		    }
	    } else {
		msg.debug("Comparing and storing actual eigenvalues");
		for (coords.t=0; coords.t<dims.t; ++coords.t)
		    for (coords.y=0; coords.y<dims.y; ++coords.y) {
			Hxx.get(coords,ahxx);
			Hxy.get(coords,ahxy);
			Hyy.get(coords,ahyy);
			for (int x=0; x<dims.x; ++x) {
			    final float b = -(ahxx[x] + ahyy[x]);
			    final float c = ahxx[x]*ahyy[x] - ahxy[x]*ahxy[x];
			    final float h1 = (float)(-0.5*(b + (b < 0.0f ? -1.0f : 1.0f)*Math.sqrt(b*b - 4*c)));
			    final float h2 = c/h1;
			    if (h1 > h2) { ahxx[x] = h1; ahyy[x] = h2; }
			    else { ahxx[x] = h2; ahyy[x] = h1; }
			}
			Hxx.set(coords,ahxx);
			Hyy.set(coords,ahyy);
			pgs.step();
		    }
	    }
	    pgs.end();

	    eigenimages = new FloatImage[2];
	    eigenimages[0] = Hxx;
	    eigenimages[1] = Hyy;
	}
	else { // 3D case

	    // Compute the components of the Hessian:
	    final boolean bDebug = msg.debug();
	    FloatImage Hxx, Hxy, Hxz, Hyy, Hyz, Hzz;
	    if (Features.sepSmooth()) {
		msg.debug("Seperating smoothing and differentiation");
		msg.debug("Gaussian smoothing at scale "+scale);
		msg.debug(false);
		final FloatImage smoothImage = useimage ? image : image.duplicate();
		differ.smooth(smoothImage,scale,false);
		msg.debug(bDebug);
		msg.debug("Computing Hxx");
		msg.debug(false);
		Hxx = differ.fdiff(smoothImage,0.0f,2,0,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Hxy");
		msg.debug(false);
		Hxy = differ.fdiff(smoothImage,0.0f,1,1,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Hxz");
		msg.debug(false);
		Hxz = differ.fdiff(smoothImage,0.0f,1,0,1,true);
		msg.debug(bDebug);
		msg.debug("Computing Hyy");
		msg.debug(false);
		Hyy = differ.fdiff(smoothImage,0.0f,0,2,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Hyz");
		msg.debug(false);
		Hyz = differ.fdiff(smoothImage,0.0f,0,1,1,true);
		msg.debug(bDebug);
		msg.debug("Computing Hzz");
		msg.debug(false);
		Hzz = differ.fdiff(smoothImage,0.0f,0,0,2,false);
		msg.debug(bDebug);
	    } else {
		msg.debug("Combining smoothing and differentiation");
		msg.debug("Gaussian differentiation at scale "+scale);
		msg.debug("Computing Hxx");
		msg.debug(false);
		Hxx = differ.gauss(image,scale,2,0,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Hxy");
		msg.debug(false);
		Hxy = differ.gauss(image,scale,1,1,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Hxz");
		msg.debug(false);
		Hxz = differ.gauss(image,scale,1,0,1,true);
		msg.debug(bDebug);
		msg.debug("Computing Hyy");
		msg.debug(false);
		Hyy = differ.gauss(image,scale,0,2,0,true);
		msg.debug(bDebug);
		msg.debug("Computing Hyz");
		msg.debug(false);
		Hyz = differ.gauss(image,scale,0,1,1,true);
		msg.debug(bDebug);
		msg.debug("Computing Hzz");
		msg.debug(false);
		Hzz = differ.gauss(image,scale,0,0,2,useimage?false:true);
		msg.debug(bDebug);
	    }

	    // Compute the eigenimages: (note that in order to minimize
	    // memory usage, Hxx, Hyy, and Hzz are re-used)
	    msg.debug("Computing 3D eigenimages");
	    final Progressor pgs = new Progressor("Computing eigenimages");
	    pgs.steps(dims.t*dims.z*dims.y);
	    Hxx.axes(Axes.X); Hxy.axes(Axes.X); Hxz.axes(Axes.X);
	    Hyy.axes(Axes.X); Hyz.axes(Axes.X); Hzz.axes(Axes.X);
	    final float[] ahxx = new float[dims.x];
	    final float[] ahxy = new float[dims.x];
	    final float[] ahxz = new float[dims.x];
	    final float[] ahyy = new float[dims.x];
	    final float[] ahyz = new float[dims.x];
	    final float[] ahzz = new float[dims.x];
	    final Coordinates coords = new Coordinates();

	    if (absolute) {
		msg.debug("Comparing and storing absolute eigenvalues");
		for (coords.t=0; coords.t<dims.t; ++coords.t)
		    for (coords.z=0; coords.z<dims.z; ++coords.z)
			for (coords.y=0; coords.y<dims.y; ++coords.y) {
			    Hxx.get(coords,ahxx);
			    Hxy.get(coords,ahxy);
			    Hxz.get(coords,ahxz);
			    Hyy.get(coords,ahyy);
			    Hyz.get(coords,ahyz);
			    Hzz.get(coords,ahzz);
			    for (int x=0; x<dims.x; ++x) {
				final float fhxx = ahxx[x];
				final float fhxy = ahxy[x];
				final float fhxz = ahxz[x];
				final float fhyy = ahyy[x];
				final float fhyz = ahyz[x];
				final float fhzz = ahzz[x];
				final float a = -(fhxx + fhyy + fhzz);
				final float b = fhxx*fhyy + fhxx*fhzz + fhyy*fhzz - fhxy*fhxy - fhxz*fhxz - fhyz*fhyz;
				final float c = fhxx*(fhyz*fhyz - fhyy*fhzz) + fhyy*fhxz*fhxz +	fhzz*fhxy*fhxy - 2*fhxy*fhxz*fhyz;
				final double q = (a*a - 3*b)/9;
				final double r = (a*a*a - 4.5f*a*b + 13.5f*c)/27f;
				final double sqrtq = (q > 0) ? Math.sqrt(q) : 0.0;
				final double rsqq3 = r/(sqrtq*sqrtq*sqrtq);
				final double angle = (rsqq3*rsqq3 <= 1) ? Math.acos(rsqq3) : Math.acos(rsqq3 < 0.0f ? -1.0f : 1.0f);
				float absh1 = (float)Math.abs(-2.0*sqrtq*Math.cos(angle/3) - a/3);
				float absh2 = (float)Math.abs(-2.0*sqrtq*Math.cos((angle + TWOPI)/3) - a/3);
				float absh3 = (float)Math.abs(-2.0*sqrtq*Math.cos((angle - TWOPI)/3) - a/3);
				if (absh2 < absh3) { final float tmp = absh2; absh2 = absh3; absh3 = tmp; }
				if (absh1 < absh2) { final float tmp1 = absh1; absh1 = absh2; absh2 = tmp1;
				if (absh2 < absh3) { final float tmp2 = absh2; absh2 = absh3; absh3 = tmp2; }}
				ahxx[x] = absh1;
				ahyy[x] = absh2;
				ahzz[x] = absh3;
			    }
			    Hxx.set(coords,ahxx);
			    Hyy.set(coords,ahyy);
			    Hzz.set(coords,ahzz);
			    pgs.step();
			}
	    } else {
		msg.debug("Comparing and storing actual eigenvalues");
		for (coords.t=0; coords.t<dims.t; ++coords.t)
		    for (coords.z=0; coords.z<dims.z; ++coords.z)
			for (coords.y=0; coords.y<dims.y; ++coords.y) {
			    Hxx.get(coords,ahxx);
			    Hxy.get(coords,ahxy);
			    Hxz.get(coords,ahxz);
			    Hyy.get(coords,ahyy);
			    Hyz.get(coords,ahyz);
			    Hzz.get(coords,ahzz);
			    for (int x=0; x<dims.x; ++x) {
				final float fhxx = ahxx[x];
				final float fhxy = ahxy[x];
				final float fhxz = ahxz[x];
				final float fhyy = ahyy[x];
				final float fhyz = ahyz[x];
				final float fhzz = ahzz[x];
				final float a = -(fhxx + fhyy + fhzz);
				final float b = fhxx*fhyy + fhxx*fhzz + fhyy*fhzz - fhxy*fhxy - fhxz*fhxz - fhyz*fhyz;
				final float c = fhxx*(fhyz*fhyz - fhyy*fhzz) + fhyy*fhxz*fhxz +	fhzz*fhxy*fhxy - 2*fhxy*fhxz*fhyz;
				final double q = (a*a - 3*b)/9;
				final double r = (a*a*a - 4.5f*a*b + 13.5f*c)/27f;
				final double sqrtq = (q > 0) ? Math.sqrt(q) : 0.0;
				final double rsqq3 = r/(sqrtq*sqrtq*sqrtq);
				final double angle = (rsqq3*rsqq3 <= 1) ? Math.acos(rsqq3) : Math.acos(rsqq3 < 0.0f ? -1.0f : 1.0f);
				float h1 = (float)(-2.0*sqrtq*Math.cos(angle/3) - a/3);
				float h2 = (float)(-2.0*sqrtq*Math.cos((angle + TWOPI)/3) - a/3);
				float h3 = (float)(-2.0*sqrtq*Math.cos((angle - TWOPI)/3) - a/3);
				if (h2 < h3) { final float tmp = h2; h2 = h3; h3 = tmp; }
				if (h1 < h2) { final float tmp1 = h1; h1 = h2; h2 = tmp1;
				if (h2 < h3) { final float tmp2 = h2; h2 = h3; h3 = tmp2; }}
				ahxx[x] = h1;
				ahyy[x] = h2;
				ahzz[x] = h3;
			    }
			    Hxx.set(coords,ahxx);
			    Hyy.set(coords,ahyy);
			    Hzz.set(coords,ahzz);
			    pgs.step();
			}
	    }
	    pgs.end();

	    eigenimages = new FloatImage[3];
	    eigenimages[0] = Hxx;
	    eigenimages[1] = Hyy;
	    eigenimages[2] = Hzz;
	}

	tmr.stop(msg);
	return eigenimages;
    }

    private final Messenger msg = Features.msg;
    private final Differentiator differ = new Differentiator();

    private final double TWOPI = 2*Math.PI;
}
