// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.features;

import imagescience.images.Axes;
import imagescience.images.ByteImage;
import imagescience.images.Coordinates;
import imagescience.images.Dimensions;
import imagescience.images.FloatImage;
import imagescience.images.ShortImage;
import imagescience.utilities.Progressor;

/** Provides methods for computing statistical features of images. */
public final class Statistics {
    
    public final static int MINIMUM=1, MAXIMUM=2, MEAN=4, MEDIAN=8,
	VOXELS=16, MASS=32, VARIANCE=64, MODE=128, SDEVIATION=256,
	ADEVIATION=512, L1NORM=1024, L2NORM=2048, SKEWNESS=4096,
	KURTOSIS=8192;
    
    public final static int NONE=0, ALL=16383;
    
    private double dMinimum, dMaximum, dMean, dMedian, dVoxels, dMass,
	dVariance, dMode, dSDeviation, dADeviation, dL1norm, dL2norm,
	dSkewness, dKurtosis;
    
    private int iStatistics = 0;
    
    private final String statusInfo = "Computing statistics";
    
    /** Method to specify the statistics that must be computed in the
	next call to <code>run</code>.

	@param statistics any or a combination (by addition) of the
	static fields of this class.
 
	@exception IllegalArgumentException if the argument is not any
	or a combination of the static fields of this class.
    */
    public synchronized void set(final int statistics) {
	iStatistics = statistics;
	if (iStatistics < NONE || iStatistics > ALL)
	    throw new IllegalArgumentException("Non-supported statistics");
    }
    
    /** Returns the value of the requested statistic as computed in the
	last call to <code>run</code>.

	@param statistics any of the static fields of this class.
 
	@return the value of the requested statistic.<br><ul> If the
	requested statistic was not computed in the last execution of
	method <code>run</code>, the returned value is
	<code>java.lang.Double.NaN</code>.</ul>

	@exception IllegalArgumentException if the argument is not any
	of the static fields of this class.
    */
    public double get(final int statistic) {
	
	switch (statistic) {
	case MINIMUM: return dMinimum;
	case MAXIMUM: return dMaximum;
	case MEAN: return dMean;
	case MEDIAN: return dMedian;
	case VOXELS: return dVoxels;
	case MASS: return dMass;
	case VARIANCE: return dVariance;
	case MODE: return dMode;
	case SDEVIATION: return dSDeviation;
	case ADEVIATION: return dADeviation;
	case L1NORM: return dL1norm;
	case L2NORM: return dL2norm;
	case SKEWNESS: return dSkewness;
	case KURTOSIS: return dKurtosis;
	default: throw new IllegalArgumentException("Non-supported statistic");
	}
    }
    
    /** Method to compute statistics of an image of type
	<code>ByteImage</code>. In principle, only those
	statistics are computed that were included in the last call to
	method <code>set</code>. The others are set to
	<code>java.lang.Double.NaN</code>. However, since some
	statistics can not be computed without knowledge of some other
	statistics (e.g. the variance requires knowledge of the mean),
	it may happen that this method computes more statistics than
	requested.

	@param image the input image whose statistics are to be computed.
 
	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public synchronized void run(final ByteImage image) {
	final Dimensions dims = image.dimensions();
	final Coordinates min = new Coordinates();
	final Coordinates max = new Coordinates(dims.x-1,dims.y-1,dims.z-1,dims.t-1);
	run(image,min,max);
    }
    
    /** Method to compute statistics within a region of interest of an
	image of type <code>ByteImage</code>. In principle, only
	those statistics are computed that were included in the last
	call to method <code>set</code>. The others are set to
	<code>java.lang.Double.NaN</code>. However, since some
	statistics can not be computed without knowledge of some other
	statistics (e.g. the variance requires knowledge of the mean),
	it may happen that this method computes more statistics than
	requested.

	@param image the input image whose statistics are to be computed.
 
	@param min,max the minimum and maximum coordinates of the ROI.
 
	@exception IllegalArgumentException if the ROI does not fall
	entirely within the image, or if any of the minimum coordinates
	is larger than its corresponding maximum coordinate.

	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public synchronized void run(final ByteImage image, final Coordinates min, final Coordinates max) {
	
	// Initialize:
	dMinimum = dMaximum = dMean = dMedian = dVoxels = dMass =
	    dVariance = dMode = dSDeviation = dADeviation = dL1norm =
	    dL2norm = dSkewness = dKurtosis = Double.NaN;
	
	final Dimensions dims = image.dimensions();
	if (min.x > max.x || min.x < 0 || max.x >= dims.x ||
	    min.y > max.y || min.y < 0 || max.y >= dims.y ||
	    min.z > max.z || min.z < 0 || max.z >= dims.z ||
	    min.t > max.t || min.t < 0 || max.t >= dims.t)
	    throw new IllegalArgumentException("Invalid region of interest");
	
	final Progressor pgs = new Progressor(statusInfo);
	final int iSteps = (max.t-min.t+1)*(max.z-min.z+1);
	final int i2Steps = 2*iSteps;
	
	final Coordinates c = new Coordinates(); c.x = min.x;
	final int xlen = max.x - min.x + 1;
	final float[] a = new float[xlen];
	image.axes(Axes.X);
	
	// Determine the types of statistics to be computed:
	final boolean bMost = (iStatistics & (MINIMUM+MAXIMUM+MEAN+VARIANCE+SDEVIATION+L1NORM+L2NORM+MASS)) != 0;
	final boolean bHist = (iStatistics & (MEDIAN+MODE)) != 0;
	final boolean bSkku = (iStatistics & (SKEWNESS+KURTOSIS)) != 0;
	final boolean bMath = (iStatistics & (ADEVIATION)) != 0;
	
	final int iCombined = (bMost ? 1 : 0) + (bHist ? 10 : 0) + (bSkku ? 100 : 0) + (bMath ? 1000 : 0);
	
	// Compute statistics: (the different types are combined so as
	// to minimize the number of passes through the data set)
	dVoxels = (max.x-min.x+1)*(max.y-min.y+1)*(max.z-min.z+1)*(max.t-min.t+1);
	
	switch (iCombined) {
	case 1: { // Most
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    dSum1 += val1;
			    dSum2 += val1*val1;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    break;
	}
	case 10: { // Hist
	    pgs.steps(iSteps);
	    final int[] hist = new int[256];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) ++hist[(int)a[x]];
		    }
		    pgs.step();
		}
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<256; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    break;
	}
	case 11: { // Most + Hist
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    final int[] hist = new int[256];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    dSum1 += val1;
			    dSum2 += val1*val1;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			    ++hist[(int)val1];
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<256; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    break;
	}
	case 100: { // Skku
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			}
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    break;
	}
	case 101: { // Most + Skku
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    break;
	}
	case 110: { // Hist + Skku
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    final int[] hist = new int[256];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    ++hist[(int)val1];
			}
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<256; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    break;
	}
	case 111: { // Most + Hist + Skku
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    final int[] hist = new int[256];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			    ++hist[(int)val1];
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<256; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    break;
	}
	case 1000: { // Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) dSum1 += a[x];
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1001: { // Most + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    dSum1 += val1;
			    dSum2 += val1*val1;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1010: { // Hist + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0;
	    final int[] hist = new int[256];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    ++hist[(int)a[x]];
			    dSum1 += a[x];
			}
		    }
		    pgs.step();
		}
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<256; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    dMean = dSum1/dVoxels;
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1011: { // Most + Hist + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    final int[] hist = new int[256];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    dSum1 += val1;
			    dSum2 += val1*val1;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			    ++hist[(int)val1];
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<256; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1100: { // Skku + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			}
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1101: { // Most + Skku + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1110: { // Hist + Skku + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    final int[] hist = new int[256];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    ++hist[(int)val1];
			}
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<256; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1111: { // Most + Hist + Skku + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    final int[] hist = new int[256];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			    ++hist[(int)val1];
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<256; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	}
	
	pgs.end();
    }

    /** Method to compute statistics of an image of type
	<code>ShortImage</code>. In principle, only those
	statistics are computed that were included in the last call to
	method <code>set</code>. The others are set to
	<code>java.lang.Double.NaN</code>. However, since some
	statistics can not be computed without knowledge of some other
	statistics (e.g. the variance requires knowledge of the mean),
	it may happen that this method computes more statistics than
	requested.

	@param image the input image whose statistics are to be computed.
 
	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public synchronized void run(final ShortImage image) {
	final Dimensions dims = image.dimensions();
	final Coordinates min = new Coordinates();
	final Coordinates max = new Coordinates(dims.x-1,dims.y-1,dims.z-1,dims.t-1);
	run(image,min,max);
    }
    
    /** Method to compute statistics within a region of interest of an
	image of type <code>ShortImage</code>. In principle, only
	those statistics are computed that were included in the last
	call to method <code>set</code>. The others are set to
	<code>java.lang.Double.NaN</code>. However, since some
	statistics can not be computed without knowledge of some other
	statistics (e.g. the variance requires knowledge of the mean),
	it may happen that this method computes more statistics than
	requested.

	@param image the input image whose statistics are to be computed.
 
	@param min,max the minimum and maximum coordinates of the ROI.
 
	@exception IllegalArgumentException if the ROI does not fall
	entirely within the image, or if any of the minimum coordinates
	is larger than its corresponding maximum coordinate.

	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public synchronized void run(final ShortImage image, final Coordinates min, final Coordinates max) {
	
	// Initialize:
	dMinimum = dMaximum = dMean = dMedian = dVoxels = dMass =
	    dVariance = dMode = dSDeviation = dADeviation = dL1norm =
	    dL2norm = dSkewness = dKurtosis = Double.NaN;
	
	final Dimensions dims = image.dimensions();
	if (min.x > max.x || min.x < 0 || max.x >= dims.x ||
	    min.y > max.y || min.y < 0 || max.y >= dims.y ||
	    min.z > max.z || min.z < 0 || max.z >= dims.z ||
	    min.t > max.t || min.t < 0 || max.t >= dims.t)
	    throw new IllegalArgumentException("Invalid region of interest");
	
	final Progressor pgs = new Progressor(statusInfo);
	final int iSteps = (max.t-min.t+1)*(max.z-min.z+1);
	final int i2Steps = 2*iSteps;
	
	final Coordinates c = new Coordinates(); c.x = min.x;
	final int xlen = max.x - min.x + 1;
	final float[] a = new float[xlen];
	image.axes(Axes.X);
	
	// Determine the types of statistics to be computed:
	final boolean bMost = (iStatistics & (MINIMUM+MAXIMUM+MEAN+VARIANCE+SDEVIATION+L1NORM+L2NORM+MASS)) != 0;
	final boolean bHist = (iStatistics & (MEDIAN+MODE)) != 0;
	final boolean bSkku = (iStatistics & (SKEWNESS+KURTOSIS)) != 0;
	final boolean bMath = (iStatistics & (ADEVIATION)) != 0;
	
	final int iCombined = (bMost ? 1 : 0) + (bHist ? 10 : 0) + (bSkku ? 100 : 0) + (bMath ? 1000 : 0);
	
	// Compute statistics: (the different types are combined so as
	// to minimize the number of passes through the data set)
	dVoxels = (max.x-min.x+1)*(max.y-min.y+1)*(max.z-min.z+1)*(max.t-min.t+1);
	
	switch (iCombined) {
	case 1: { // Most
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    dSum1 += val1;
			    dSum2 += val1*val1;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    break;
	}
	case 10: { // Hist
	    pgs.steps(iSteps);
	    final int[] hist = new int[65536];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) ++hist[(int)a[x]];
		    }
		    pgs.step();
		}
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<65536; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    break;
	}
	case 11: { // Most + Hist
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    final int[] hist = new int[65536];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    dSum1 += val1;
			    dSum2 += val1*val1;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			    ++hist[(int)val1];
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<65536; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    break;
	}
	case 100: { // Skku
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			}
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    break;
	}
	case 101: { // Most + Skku
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    break;
	}
	case 110: { // Hist + Skku
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    final int[] hist = new int[65536];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    ++hist[(int)val1];
			}
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<65536; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    break;
	}
	case 111: { // Most + Hist + Skku
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    final int[] hist = new int[65536];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			    ++hist[(int)val1];
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<65536; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    break;
	}
	case 1000: { // Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) dSum1 += a[x];
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1001: { // Most + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    dSum1 += val1;
			    dSum2 += val1*val1;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1010: { // Hist + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0;
	    final int[] hist = new int[65536];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    ++hist[(int)a[x]];
			    dSum1 += a[x];
			}
		    }
		    pgs.step();
		}
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<65536; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    dMean = dSum1/dVoxels;
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1011: { // Most + Hist + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    final int[] hist = new int[65536];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    dSum1 += val1;
			    dSum2 += val1*val1;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			    ++hist[(int)val1];
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<65536; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1100: { // Skku + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			}
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1101: { // Most + Skku + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1110: { // Hist + Skku + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    final int[] hist = new int[65536];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    ++hist[(int)val1];
			}
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<65536; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1111: { // Most + Hist + Skku + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    final int[] hist = new int[65536];
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			    ++hist[(int)val1];
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    dL1norm = dMass; // in this case of always positive numbers
	    dL2norm = Math.sqrt(dSum2);
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    int iMax = 0, iCum = 0;
	    final int iVoxels = (int)dVoxels;
	    final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
	    for (int i=0; i<65536; ++i) {
		final int iVal = hist[i];
		if (iVal >= iMax) { iMax = iVal; dMode = i; }
		if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
	    }
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	}
	
	pgs.end();
    }

    /** Method to compute statistics of an image of type
	<code>FloatImage</code>. In principle, only those
	statistics are computed that were included in the last call to
	method <code>set</code>. The others are set to
	<code>java.lang.Double.NaN</code>. However, since some
	statistics can not be computed without knowledge of some other
	statistics (e.g. the variance requires knowledge of the mean),
	it may happen that this method computes more statistics than
	requested. Also note that the median and mode are not computed
	exactly but are estimated with an accuracy of +/- 0.005 percent
	of the dynamic range of the image.

	@param image the input image whose statistics are to be computed.
 
	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public synchronized void run(final FloatImage image) {
	final Dimensions dims = image.dimensions();
	final Coordinates min = new Coordinates();
	final Coordinates max = new Coordinates(dims.x-1,dims.y-1,dims.z-1,dims.t-1);
	run(image,min,max);
    }
    
    /** Method to compute statistics within a region of interest of an
	image of type <code>FloatImage</code>. In principle, only
	those statistics are computed that were included in the last
	call to method <code>set</code>. The others are set to
	<code>java.lang.Double.NaN</code>. However, since some
	statistics can not be computed without knowledge of some other
	statistics (e.g. the variance requires knowledge of the mean),
	it may happen that this method computes more statistics than
	requested. Also note that the median and mode are not computed
	exactly but are estimated with an accuracy of +/- 0.005 percent
	of the dynamic range of the image.

	@param image the input image whose statistics are to be computed.
 
	@param min,max the minimum and maximum coordinates of the ROI.
 
	@exception IllegalArgumentException if the ROI does not fall
	entirely within the image, or if any of the minimum coordinates
	is larger than its corresponding maximum coordinate.

	@exception NullPointerException if the input image is
	<code>null</code>.
    */
    public synchronized void run(final FloatImage image, final Coordinates min, final Coordinates max) {
	
	// Initialize:
	dMinimum = dMaximum = dMean = dMedian = dVoxels = dMass =
	    dVariance = dMode = dSDeviation = dADeviation = dL1norm =
	    dL2norm = dSkewness = dKurtosis = Double.NaN;
	
	final Dimensions dims = image.dimensions();
	if (min.x > max.x || min.x < 0 || max.x >= dims.x ||
	    min.y > max.y || min.y < 0 || max.y >= dims.y ||
	    min.z > max.z || min.z < 0 || max.z >= dims.z ||
	    min.t > max.t || min.t < 0 || max.t >= dims.t)
	    throw new IllegalArgumentException("Invalid region of interest");
	
	final Progressor pgs = new Progressor(statusInfo);
	final int iSteps = (max.t-min.t+1)*(max.z-min.z+1);
	final int i2Steps = 2*iSteps;
	
	final int iNBins = 10000;
	final Coordinates c = new Coordinates(); c.x = min.x;
	final int xlen = max.x - min.x + 1;
	final float[] a = new float[xlen];
	image.axes(Axes.X);
	
	// Determine the types of statistics to be computed:
	final boolean bMost = (iStatistics & (MINIMUM+MAXIMUM+MEAN+VARIANCE+SDEVIATION+L1NORM+L2NORM+MASS)) != 0;
	final boolean bHist = (iStatistics & (MEDIAN+MODE)) != 0;
	final boolean bSkku = (iStatistics & (SKEWNESS+KURTOSIS)) != 0;
	final boolean bMath = (iStatistics & (ADEVIATION)) != 0;
	
	final int iCombined = (bMost ? 1 : 0) + (bHist ? 10 : 0) + (bSkku ? 100 : 0) + (bMath ? 1000 : 0);
	
	// Compute statistics: (the different types are combined so as
	// to minimize the number of passes through the data set)
	dVoxels = (max.x-min.x+1)*(max.y-min.y+1)*(max.z-min.z+1)*(max.t-min.t+1);
	
	switch (iCombined) {
	case 1: { // Most
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0, dASum=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    dSum1 += val1;
			    dSum2 += val1*val1;
			    dASum += Math.abs(val1);
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dASum;
	    dL2norm = Math.sqrt(dSum2);
	    break;
	}
	case 10: { // Hist
	    pgs.steps(i2Steps);
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    final double dRange = dMaximum - dMinimum;
	    if (dRange == 0.0)
		dMedian = dMode = dMinimum;
	    else {
		final int[] hist = new int[iNBins+1];
		final double dNBins = iNBins;
		for (c.t=min.t; c.t<=max.t; ++c.t)
		    for (c.z=min.z; c.z<=max.z; ++c.z) {
			for (c.y=min.y; c.y<=max.y; ++c.y) {
			    image.get(c,a);
			    for (int x=0; x<xlen; ++x)
				++hist[(int)(((a[x] - dMinimum)/dRange)*dNBins)];
			}
			pgs.step();
		}
		hist[iNBins-1] += hist[iNBins];
		int iMax = 0, iCum = 0;
		final int iVoxels = (int)dVoxels;
		final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
		for (int i=0; i<iNBins; ++i) {
		    final int iVal = hist[i];
		    if (iVal >= iMax) { iMax = iVal; dMode = i; }
		    if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
		}
		dMode = dMinimum + ((dMode + 0.5)/dNBins)*dRange;
		dMedian = dMinimum + ((dMedian + 0.5)/dNBins)*dRange;
	    }
	    break;
	}
	case 11: { // Most + Hist
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dASum=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    dSum1 += val1;
			    dSum2 += val1*val1;
			    dASum += Math.abs(val1);
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dASum;
	    dL2norm = Math.sqrt(dSum2);
	    final double dRange = dMaximum - dMinimum;
	    if (dRange == 0.0)
		dMedian = dMode = dMinimum;
	    else {
		final int[] hist = new int[iNBins+1];
		final double dNBins = iNBins;
		for (c.t=min.t; c.t<=max.t; ++c.t)
		    for (c.z=min.z; c.z<=max.z; ++c.z) {
			for (c.y=min.y; c.y<=max.y; ++c.y) {
			    image.get(c,a);
			    for (int x=0; x<xlen; ++x)
				++hist[(int)(((a[x] - dMinimum)/dRange)*dNBins)];
			}
			pgs.step();
		}
		hist[iNBins-1] += hist[iNBins];
		int iMax = 0, iCum = 0;
		final int iVoxels = (int)dVoxels;
		final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
		for (int i=0; i<iNBins; ++i) {
		    final int iVal = hist[i];
		    if (iVal >= iMax) { iMax = iVal; dMode = i; }
		    if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
		}
		dMode = dMinimum + ((dMode + 0.5)/dNBins)*dRange;
		dMedian = dMinimum + ((dMedian + 0.5)/dNBins)*dRange;
	    }
	    break;
	}
	case 100: { // Skku
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			}
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    break;
	}
	case 101: { // Most + Skku
	    pgs.steps(iSteps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0, dASum=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    dASum += Math.abs(val1);
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dASum;
	    dL2norm = Math.sqrt(dSum2);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    break;
	}
	case 110: { // Hist + Skku
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    final double dRange = dMaximum - dMinimum;
	    if (dRange == 0.0)
		dMedian = dMode = dMinimum;
	    else {
		final int[] hist = new int[iNBins+1];
		final double dNBins = iNBins;
		for (c.t=min.t; c.t<=max.t; ++c.t)
		    for (c.z=min.z; c.z<=max.z; ++c.z) {
			for (c.y=min.y; c.y<=max.y; ++c.y) {
			    image.get(c,a);
			    for (int x=0; x<xlen; ++x)
				++hist[(int)(((a[x] - dMinimum)/dRange)*dNBins)];
			}
			pgs.step();
		}
		hist[iNBins-1] += hist[iNBins];
		int iMax = 0, iCum = 0;
		final int iVoxels = (int)dVoxels;
		final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
		for (int i=0; i<iNBins; ++i) {
		    final int iVal = hist[i];
		    if (iVal >= iMax) { iMax = iVal; dMode = i; }
		    if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
		}
		dMode = dMinimum + ((dMode + 0.5)/dNBins)*dRange;
		dMedian = dMinimum + ((dMedian + 0.5)/dNBins)*dRange;
	    }
	    break;
	}
	case 111: { // Most + Hist + Skku
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0, dASum=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    dASum += Math.abs(val1);
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dASum;
	    dL2norm = Math.sqrt(dSum2);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    final double dRange = dMaximum - dMinimum;
	    if (dRange == 0.0)
		dMedian = dMode = dMinimum;
	    else {
		final int[] hist = new int[iNBins+1];
		final double dNBins = iNBins;
		for (c.t=min.t; c.t<=max.t; ++c.t)
		    for (c.z=min.z; c.z<=max.z; ++c.z) {
			for (c.y=min.y; c.y<=max.y; ++c.y) {
			    image.get(c,a);
			    for (int x=0; x<xlen; ++x)
				++hist[(int)(((a[x] - dMinimum)/dRange)*dNBins)];
			}
			pgs.step();
		}
		hist[iNBins-1] += hist[iNBins];
		int iMax = 0, iCum = 0;
		final int iVoxels = (int)dVoxels;
		final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
		for (int i=0; i<iNBins; ++i) {
		    final int iVal = hist[i];
		    if (iVal >= iMax) { iMax = iVal; dMode = i; }
		    if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
		}
		dMode = dMinimum + ((dMode + 0.5)/dNBins)*dRange;
		dMedian = dMinimum + ((dMedian + 0.5)/dNBins)*dRange;
	    }
	    break;
	}
	case 1000: { // Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) dSum1 += a[x];
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1001: { // Most + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dASum=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    dSum1 += val1;
			    dSum2 += val1*val1;
			    dASum += Math.abs(val1);
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dASum;
	    dL2norm = Math.sqrt(dSum2);
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1010: { // Hist + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    dSum1 += val1;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    double dADevSum=0.0;
	    final double dRange = dMaximum - dMinimum;
	    if (dRange == 0.0) {
		dMedian = dMode = dMinimum;
		for (c.t=min.t; c.t<=max.t; ++c.t)
		    for (c.z=min.z; c.z<=max.z; ++c.z) {
			for (c.y=min.y; c.y<=max.y; ++c.y) {
			    image.get(c,a);
			    for (int x=0; x<xlen; ++x)
				dADevSum += Math.abs(a[x] - dMean);
			}
			pgs.step();
		    }
		dADeviation = dADevSum/dVoxels;
	    }
	    else {
		final int[] hist = new int[iNBins+1];
		final double dNBins = iNBins;
		for (c.t=min.t; c.t<=max.t; ++c.t)
		    for (c.z=min.z; c.z<=max.z; ++c.z) {
			for (c.y=min.y; c.y<=max.y; ++c.y) {
			    image.get(c,a);
			    for (int x=0; x<xlen; ++x) {
				++hist[(int)(((a[x] - dMinimum)/dRange)*dNBins)];
				dADevSum += Math.abs(a[x] - dMean);
			    }
			}
			pgs.step();
		    }
		hist[iNBins-1] += hist[iNBins];
		int iMax = 0, iCum = 0;
		final int iVoxels = (int)dVoxels;
		final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
		for (int i=0; i<iNBins; ++i) {
		    final int iVal = hist[i];
		    if (iVal >= iMax) { iMax = iVal; dMode = i; }
		    if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
		}
		dMode = dMinimum + ((dMode + 0.5)/dNBins)*dRange;
		dMedian = dMinimum + ((dMedian + 0.5)/dNBins)*dRange;
		dADeviation = dADevSum/dVoxels;
	    }
	    break;
	}
	case 1011: { // Most + Hist + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dASum=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    dSum1 += val1;
			    dSum2 += val1*val1;
			    dASum += Math.abs(val1);
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dASum;
	    dL2norm = Math.sqrt(dSum2);
	    double dADevSum=0.0;
	    final double dRange = dMaximum - dMinimum;
	    if (dRange == 0.0) {
		dMedian = dMode = dMinimum;
		for (c.t=min.t; c.t<=max.t; ++c.t)
		    for (c.z=min.z; c.z<=max.z; ++c.z) {
			for (c.y=min.y; c.y<=max.y; ++c.y) {
			    image.get(c,a);
			    for (int x=0; x<xlen; ++x)
				dADevSum += Math.abs(a[x] - dMean);
			}
			pgs.step();
		    }
		dADeviation = dADevSum/dVoxels;
	    }
	    else {
		final int[] hist = new int[iNBins+1];
		final double dNBins = iNBins;
		for (c.t=min.t; c.t<=max.t; ++c.t)
		    for (c.z=min.z; c.z<=max.z; ++c.z) {
			for (c.y=min.y; c.y<=max.y; ++c.y) {
			    image.get(c,a);
			    for (int x=0; x<xlen; ++x) {
				++hist[(int)(((a[x] - dMinimum)/dRange)*dNBins)];
				dADevSum += Math.abs(a[x] - dMean);
			    }
			}
			pgs.step();
		    }
		hist[iNBins-1] += hist[iNBins];
		int iMax = 0, iCum = 0;
		final int iVoxels = (int)dVoxels;
		final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
		for (int i=0; i<iNBins; ++i) {
		    final int iVal = hist[i];
		    if (iVal >= iMax) { iMax = iVal; dMode = i; }
		    if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
		}
		dMode = dMinimum + ((dMode + 0.5)/dNBins)*dRange;
		dMedian = dMinimum + ((dMedian + 0.5)/dNBins)*dRange;
		dADeviation = dADevSum/dVoxels;
	    }
	    break;
	}
	case 1100: { // Skku + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			}
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1101: { // Most + Skku + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0, dASum=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    dASum += Math.abs(val1);
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dASum;
	    dL2norm = Math.sqrt(dSum2);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    double dADevSum=0.0;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x)
			    dADevSum += Math.abs(a[x] - dMean);
		    }
		    pgs.step();
		}
	    dADeviation = dADevSum/dVoxels;
	    break;
	}
	case 1110: { // Hist + Skku + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    double dADevSum=0.0;
	    final double dRange = dMaximum - dMinimum;
	    if (dRange == 0.0) {
		dMedian = dMode = dMinimum;
		for (c.t=min.t; c.t<=max.t; ++c.t)
		    for (c.z=min.z; c.z<=max.z; ++c.z) {
			for (c.y=min.y; c.y<=max.y; ++c.y) {
			    image.get(c,a);
			    for (int x=0; x<xlen; ++x)
				dADevSum += Math.abs(a[x] - dMean);
			}
			pgs.step();
		    }
		dADeviation = dADevSum/dVoxels;
	    }
	    else {
		final int[] hist = new int[iNBins+1];
		final double dNBins = iNBins;
		for (c.t=min.t; c.t<=max.t; ++c.t)
		    for (c.z=min.z; c.z<=max.z; ++c.z) {
			for (c.y=min.y; c.y<=max.y; ++c.y) {
			    image.get(c,a);
			    for (int x=0; x<xlen; ++x) {
				++hist[(int)(((a[x] - dMinimum)/dRange)*dNBins)];
				dADevSum += Math.abs(a[x] - dMean);
			    }
			}
			pgs.step();
		    }
		hist[iNBins-1] += hist[iNBins];
		int iMax = 0, iCum = 0;
		final int iVoxels = (int)dVoxels;
		final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
		for (int i=0; i<iNBins; ++i) {
		    final int iVal = hist[i];
		    if (iVal >= iMax) { iMax = iVal; dMode = i; }
		    if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
		}
		dMode = dMinimum + ((dMode + 0.5)/dNBins)*dRange;
		dMedian = dMinimum + ((dMedian + 0.5)/dNBins)*dRange;
		dADeviation = dADevSum/dVoxels;
	    }
	    break;
	}
	case 1111: { // Most + Hist + Skku + Math
	    pgs.steps(i2Steps);
	    double dSum1=0.0, dSum2=0.0, dSum3=0.0, dSum4=0.0, dASum=0.0;
	    dMinimum = Double.MAX_VALUE;
	    dMaximum = -Double.MAX_VALUE;
	    for (c.t=min.t; c.t<=max.t; ++c.t)
		for (c.z=min.z; c.z<=max.z; ++c.z) {
		    for (c.y=min.y; c.y<=max.y; ++c.y) {
			image.get(c,a);
			for (int x=0; x<xlen; ++x) {
			    final double val1 = a[x];
			    final double val2 = val1*val1;
			    dSum1 += val1;
			    dSum2 += val2;
			    dSum3 += val1*val2;
			    dSum4 += val2*val2;
			    dASum += Math.abs(val1);
			    if (val1 < dMinimum) dMinimum = val1;
			    if (val1 > dMaximum) dMaximum = val1;
			}
		    }
		    pgs.step();
		}
	    dMass = dSum1;
	    dMean = dSum1/dVoxels;
	    final double dMean2 = dMean*dMean;
	    dVariance = (dSum2/dVoxels) - dMean2;
	    dSDeviation = Math.sqrt(dVariance);
	    dL1norm = dASum;
	    dL2norm = Math.sqrt(dSum2);
	    dSkewness = ((dSum3 - 3.0*dMean*dSum2)/dVoxels + 2.0*dMean*dMean2)/(dVariance*dSDeviation);
	    dKurtosis = (((dSum4 - 4.0*dMean*dSum3 + 6.0*dMean2*dSum2)/dVoxels - 3.0*dMean2*dMean2)/(dVariance*dVariance)-3.0);
	    double dADevSum=0.0;
	    final double dRange = dMaximum - dMinimum;
	    if (dRange == 0.0) {
		dMedian = dMode = dMinimum;
		for (c.t=min.t; c.t<=max.t; ++c.t)
		    for (c.z=min.z; c.z<=max.z; ++c.z) {
			for (c.y=min.y; c.y<=max.y; ++c.y) {
			    image.get(c,a);
			    for (int x=0; x<xlen; ++x)
				dADevSum += Math.abs(a[x] - dMean);
			}
			pgs.step();
		    }
		dADeviation = dADevSum/dVoxels;
	    }
	    else {
		final int[] hist = new int[iNBins+1];
		final double dNBins = iNBins;
		for (c.t=min.t; c.t<=max.t; ++c.t)
		    for (c.z=min.z; c.z<=max.z; ++c.z) {
			for (c.y=min.y; c.y<=max.y; ++c.y) {
			    image.get(c,a);
			    for (int x=0; x<xlen; ++x) {
				++hist[(int)(((a[x] - dMinimum)/dRange)*dNBins)];
				dADevSum += Math.abs(a[x] - dMean);
			    }
			}
			pgs.step();
		    }
		hist[iNBins-1] += hist[iNBins];
		int iMax = 0, iCum = 0;
		final int iVoxels = (int)dVoxels;
		final int iHalfway = (iVoxels%2 == 0) ? (iVoxels/2) : (iVoxels/2 + 1);
		for (int i=0; i<iNBins; ++i) {
		    final int iVal = hist[i];
		    if (iVal >= iMax) { iMax = iVal; dMode = i; }
		    if (iCum < iHalfway) { dMedian = i; iCum += iVal; }
		}
		dMode = dMinimum + ((dMode + 0.5)/dNBins)*dRange;
		dMedian = dMinimum + ((dMedian + 0.5)/dNBins)*dRange;
		dADeviation = dADevSum/dVoxels;
	    }
	    break;
	}
	}
	
	pgs.end();
    }

}
