// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.images;

/** Contains information about the activity of the coordinate axes of
    (at most) 4D data objects. */
public final class Axes {

    /** The x-axis. */
    public final static int X=1;

    /** The y-axis. */
    public final static int Y=2;

    /** The z-axis. */
    public final static int Z=4;

    /** The t-axis. */
    public final static int T=8;

    final static int NONE=0, YX=3, ZX=5, ZY=6, TX=9, TY=10, TZ=12,
	ZYX=7, TYX=11, TZX=13, TZY=14, TZYX=15, ALL=15;

    /** Indicates whether or not the x-axis is active. */
    public boolean x = false;

    /** Indicates whether or not the y-axis is active. */
    public boolean y = false;

    /** Indicates whether or not the z-axis is active. */
    public boolean z = false;

    /** Indicates whether or not the t-axis is active. */
    public boolean t = false;

    /** Default constructor. Each of the axes is set to <tt>false</tt>. */
    public Axes() {}

    /** One-dimensional constructor. The x-axis is set to the specified
        value. The y-, z-, and t-axes are set to <tt>false</tt>. */
    public Axes(final boolean x) {
	this.x = x;
    }

    /** Two-dimensional constructor. The x- and y-axes are set to the
        specified values. The z- and t-axes are set to <tt>false</tt>. */
    public Axes(final boolean x, final boolean y) {
	this.x = x;
	this.y = y;
    }

    /** Three-dimensional constructor. The x-, y-, and z-axes are set to
        the specified values. The t-dimension is set to <tt>false</tt>. */
    public Axes(final boolean x, final boolean y, final boolean z) {
	this.x = x;
	this.y = y;
	this.z = z;
    }

    /** Four-dimensional constructor. The x-, y-, z-, and t-axes
        are set to the specified values. */
    public Axes(final boolean x, final boolean y, final boolean z, final boolean t) {
	this.x = x;
	this.y = y;
	this.z = z;
	this.t = t;
    }

    /** Four-dimensional constructor. Sets each of the four boolean
	fields according to the value of the argument, which must be one
	or a combination (by addition) of <tt>Axes.X</tt>,
	<tt>Axes.Y</tt>, <tt>Axes.Z</tt>, <tt>Axes.T</tt>.

	@exception IllegalArgumentException if the value of parameter
	<tt>axes</tt> does not correspond to a valid combination of
	axes.  */
    public Axes(final int axes) {
	switch (axes) {
	case X: x = true; break;
	case Y: y = true; break;
	case Z: z = true; break;
	case T: t = true; break;
	case YX: x = true; y = true; break;
	case ZX: x = true; z = true; break;
	case ZY: y = true; z = true; break;
	case TX: x = true; t = true; break;
	case TY: y = true; t = true; break;
	case TZ: z = true; t = true; break;
	case ZYX: x = true; y = true; z = true; break;
	case TYX: x = true; y = true; t = true; break;
	case TZX: x = true; z = true; t = true; break;
	case TZY: y = true; z = true; t = true; break;
	case TZYX: x = true; y = true; z = true; t = true; break;
	default: throw new IllegalArgumentException("Non-supported (combination of) axes");
	}
    }

    /** Returns a copy of the current <tt>Axes</tt> object. */
    public Axes duplicate() { return new Axes(x,y,z,t); }

}
