// **********************************************************************
// COPYRIGHT NOTICE AND DISCLAIMER
//
// Copyright (c) Erik Meijering. All rights reserved.
//
// Permission to copy and use this software and accompanying
// documentation for educational, research, and not-for-profit
// purposes, without fee and without a signed licensing agreement,
// is hereby granted, provided that the above copyright notice, this
// paragraph and the following two paragraphs appear in all copies.
// The copyright holder is free to make upgraded or improved versions
// of the software available for a fee or commercially only.
//
// IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
// DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// OF ANY KIND WHATSOEVER, ARISING OUT OF THE USE OF THIS SOFTWARE
// AND ITS DOCUMENTATION, EVEN IF HE HAS BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
// SOFTWARE AND ACCOMPANYING DOCUMENTATION IS PROVIDED "AS IS".
// THE COPYRIGHT HOLDER HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// **********************************************************************

package imagescience.utilities;

import ij.IJ;
import ij.gui.MessageDialog;

/** Provides methods for displaying different types of messages (error,
    debug, status, normal). The advantage of using this intermediate
    class instead of ImageJ's equivalents is that if any of the packages
    that make use of this class ever needs to be converted to some other
    java-based environment, only the methods in this class need to be
    rewritten. */
public final class Messenger {

    private boolean bDebug = false;

    /** Indicates whether or not debug messages are displayed. */
    public boolean debug() { return bDebug; }

    /** Method to enable or disable the displaying of debug messages. */
    public void debug(final boolean show) { bDebug = show; }

    /** Method for displaying debug messages. If displaying has been
        enabled for this type of messages, the text in the argument is
        written to ImageJ's results window. */
    public void debug(final String message) { if (bDebug) IJ.write(message + "\n"); }

    /** Method for displaying error messages. Currently, a call to this
        method results in a window with title "Error" showing the given
        message. It also resets ImageJ's status and progress bars. */
    public void error(final String message) {
	new MessageDialog(IJ.getInstance(), "Error", message);
	IJ.showProgress(1.0);
	IJ.showStatus("");
    }

    /** Method for displaying error messages. Currently, a call to this
        method results in a window with the given title showing the
        given message. It also resets ImageJ's status and progress bars. */
    public void error(final String title, final String message) {
	new MessageDialog(IJ.getInstance(), title, message);
	IJ.showProgress(1.0);
	IJ.showStatus("");
    }

    /** Method for displaying normal messages. Currently, a call to this
        method results in a message written to ImageJ's results window. */
    public void write(final String message) { IJ.write(message + "\n"); }

    /** Method for displaying status messages. Currently, a call to this
        method results in a message written to ImageJ's status bar. */
    public void status(final String message) { IJ.showStatus(message); }

}
